// Copyright 2022 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

// Copyright 2022 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package configuration

import (
	"encoding/json"
	"errors"
	"gopkg.in/yaml.v3"
	"strings"
)

type BoolValuer interface {
	IsTrue() bool
	IsFalse() bool
	ToBool() bool
	IsValid() bool
	String() string
	Raw() string
}

// BoolValue represents a boolean value stored as a string
type BoolValue struct {
	value bool // Holds the value bool value
	valid bool // Tracks if the value is valid ("true" or "false")
	raw   string
}

// NewBoolValue creates a new BoolValue and validates it
func NewBoolValue(raw string) (BoolValue, error) {
	bv := BoolValue{raw: raw}
	err := bv.parseAndValidate()
	return bv, err
}

func NewTrueBoolValue() BoolValue {
	return BoolValue{value: true, valid: true, raw: "true"}
}

func NewFalseBoolValue() BoolValue {
	return BoolValue{value: false, valid: true, raw: "false"}
}

// parseAndValidate processes and validates the raw string
func (p *BoolValue) parseAndValidate() error {
	switch strings.ToLower(p.raw) {
	case "true", "yes", "1", "on":
		p.value = true
		p.valid = true
	case "false", "no", "0", "off":
		p.value = false
		p.valid = true
	default:
		p.valid = false
		return errors.New("invalid BoolValue: must be 'true' or 'false'")
	}
	return nil
}

// IsTrue returns true if the value represents a logical true
func (p BoolValue) IsTrue() bool {
	return p.value && p.valid
}

// IsFalse returns true if the value represents a logical false
func (p BoolValue) IsFalse() bool {
	return !p.IsTrue()
}

// ToBool converts the BoolValue to a native bool type
func (p BoolValue) ToBool() bool {
	return p.IsTrue()
}

// IsValid checks whether the value is valid
func (p BoolValue) IsValid() bool {
	return p.valid
}

// FromBool converts a native bool to a BoolValue
func FromBool(b bool) BoolValue {
	if b {
		return BoolValue{value: true, valid: true, raw: "true"}
	}
	return BoolValue{value: false, valid: true, raw: "false"}
}

// String returns the raw string value of BoolValue
func (p BoolValue) Raw() string {
	return p.raw
}

// String returns either "true" or "false" or if the value is invalid, the raw value
func (p BoolValue) String() string {
	if p.valid {
		if p.value {
			return "true"
		}
		return "false"
	}

	return p.raw
}

// MarshalJSON serializes the BoolValue to JSON
func (p BoolValue) MarshalJSON() ([]byte, error) {
	return json.Marshal(p.IsTrue())
}

// UnmarshalJSON deserializes the BoolValue from JSON
func (p *BoolValue) UnmarshalJSON(data []byte) error {
	var b string
	var err error
	if err = json.Unmarshal(data, &b); err != nil {
		return err
	}
	*p, err = NewBoolValue(b)
	if err != nil {
		return err
	}
	return nil
}

// MarshalYAML serializes the BoolValue to YAML
func (p BoolValue) MarshalYAML() (interface{}, error) {
	return p.IsTrue(), nil
}

// UnmarshalYAML deserializes the BoolValue from YAML
func (p *BoolValue) UnmarshalYAML(value *yaml.Node) error {
	var b string
	var err error
	if err = value.Decode(&b); err != nil {
		return err
	}
	*p, err = NewBoolValue(b)
	if err != nil {
		return err
	}
	return nil
}

// MarshalTOML serializes the BoolValue to TOML
func (p BoolValue) MarshalTOML() ([]byte, error) {
	return []byte(p.raw), nil
}

// UnmarshalTOML deserializes the BoolValue from TOML
func (p *BoolValue) UnmarshalTOML(data []byte) error {
	p.raw = strings.Trim(string(data), `"`)
	return p.parseAndValidate()
}
